from PyQt5.QtWidgets import QDialog

import snitch

from .about_ui import Ui_AboutDialog

class AboutDialog(QDialog):
    """ Information dialog """
    def __init__(self, parent=None):
        super().__init__(parent)

        self._ui = Ui_AboutDialog()
        self._ui.setupUi(self)

        about_string = self._ui.text.text()
        self._ui.text.setText(about_string.replace('{VERSION_NUMBER}', snitch.__version__))



""" Application entry point """
import logging as log
import json
import pickle
import os

from PyQt5.QtCore import Qt, pyqtSlot, QModelIndex, QTimer
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import (
    QApplication, QMainWindow,
    QLabel, QFrame, QMenu, QAction,
    QFileDialog, QMessageBox, QHeaderView)
import pyautogui
from pynput.keyboard import Key

from ..model.data.testcase import TestCase
from ..model.data.events import Event
from ..model.data.screenshot import Screenshot
from ..model.eventproperties import EventPropertiesModel, EventPropertiesItemDelegate
from ..model.eventlist import EventListModel
from ..model.properties import PropertiesTableModel
from ..model.snaplist import SnapListModel
from .controller_ui import Ui_Controller
from .about import AboutDialog
from .settings import SettingsDialog
from ..recorder import EventRecorder, KeyEventCatcher
from ..player import EventPlayer, SnapshotPlayer
from ..settings import SETTINGS, CFG
from ..snapper import Snapper

class Data:
    #pylint: disable=too-few-public-methods
    TYPES = [pickle, json]
    NAMES = ["Pickle (*.p)", "Json (*.json *.js)"]
    DEFAULT_TYPE = json
    PANES_SIZES = [692, 308]

class Controller(QMainWindow):
    STATUS_TEXT = QApplication.translate('Controller', '{0} events recorded.')
    """ Main window for the app """
    def __init__(self, parent=None):
        super().__init__(parent)

        self._ui = Ui_Controller()
        self._ui.setupUi(self)

        self.new_testcase()

        self.update_open_recent_menu()

        self._ui.tableEvents.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self._ui.eventSplitter.setSizes(Data.PANES_SIZES)
        self._ui.snapshotSplitter.setSizes(Data.PANES_SIZES)

        self._selected_event = None
        self._selected_snap = None
        self._saved_geometry = None
        self._catch_interrupt = None

        self.snap = None

        self._ui.menuAdd = QMenu(self._ui.eventMgmtButtons.ui.buttonAdd)
        for event_type in Event.get_subclasses():
            self._ui.menuAdd.addAction(event_type.__name__)
        self._ui.menuAdd.triggered.connect(self.add_event)
        self._ui.eventMgmtButtons.ui.buttonAdd.setMenu(self._ui.menuAdd)

        self._ui.tableEvents.selectionModel().selectionChanged.connect(self.update_selected_event)
        self._ui.tableProperties.setModel(EventPropertiesModel(None, self))
        self._ui.tableProperties.setItemDelegate(EventPropertiesItemDelegate(self))

        self._ui.listSnap.selectionModel().selectionChanged.connect(self.update_selected_snap)
        self._ui.tableSnapProps.setModel(PropertiesTableModel(None, self))

        self._ui.eventMgmtButtons.ui.buttonUp.clicked.connect(self.move_up)
        self._ui.eventMgmtButtons.ui.buttonDown.clicked.connect(self.move_down)

        self._ui.capsEventMgmtButtons.ui.buttonUp.clicked.connect(self.move_snap_up)
        self._ui.capsEventMgmtButtons.ui.buttonDown.clicked.connect(self.move_snap_down)

        self._ui.buttonPlay.clicked.connect(self.playback)
        self._ui.capsEventMgmtButtons.ui.buttonAdd.clicked.connect(self.capture)
        self._ui.buttonCapture.clicked.connect(self.capture)
        self._ui.buttonDiff.clicked.connect(self.diff)
        self._ui.buttonOCR.clicked.connect(self.ocr)

        self._ui.statuspos = QLabel(self)
        self._ui.statuspos.setMinimumWidth(120)
        self._ui.statuspos.setFrameShape(QFrame.Panel)
        self._ui.statuspos.setFrameShadow(QFrame.Sunken)
        self._ui.statuspos.setAlignment(Qt.AlignCenter)
        self._ui.statusbar.addPermanentWidget(self._ui.statuspos)

        self._mousepos_timer = QTimer(self)
        self._mousepos_timer.timeout.connect(self.update_mouse_pos)
        self._mousepos_timer.start(50)


######## ##     ## ######## ##    ## ########  ######
##       ##     ## ##       ###   ##    ##    ##    ##
##       ##     ## ##       ####  ##    ##    ##
######   ##     ## ######   ## ## ##    ##     ######
##        ##   ##  ##       ##  ####    ##          ##
##         ## ##   ##       ##   ###    ##    ##    ##
########    ###    ######## ##    ##    ##     ######


    def select_event(self, event):
        """ Selects the row in the table for the event passed as parameter """
        self._ui.tableEvents.selectRow(self._testcase.event_list.events.index(event))
        self._ui.tableEvents.scrollTo(self._ui.tableEvents.selectedIndexes()[0])
        # Allows UI update, but may be time consuming, remove if necessary:
        QApplication.processEvents()

    def get_selected_index(self):
        """
        :return: The index of the selected row of the event list,
                 or -1 if the selection is empty
        """
        selected_rows = self._ui.tableEvents.selectionModel().selectedRows()
        if selected_rows:
            # this is ok since the selection mode/behavior is single/rows
            return selected_rows[0].row()
        return -1

    @pyqtSlot()
    def playback(self, include_snapshots=False):
        """ Plays the recorded events """
        keys = [{Key.esc}]
        with KeyEventCatcher(keys, self._player.stop) as catcher:
            self._player.play(
                start_at=self.get_selected_index(),
                event_played_callback=self.select_event,
                key_event_catcher=catcher
            )
        if include_snapshots:
            return self._snap_player.play()

        QMessageBox.information(
            self,
            self.tr("Information"),
            self.tr("Sequence playback over.")
        )
        self._ui.tableEvents.clearSelection()

    @pyqtSlot()
    def stop_playback(self):
        """ Interrupts the playback """
        self._player.stop()


    @pyqtSlot(QAction)
    def add_event(self, event_type):
        """
        :type event_type: QAction
        """
        self._testcase.event_list.insert(self.get_selected_index(), event_type.text())

    @pyqtSlot()
    def move_up(self):
        """ Move selected event up one row """
        self._move(-1)

    @pyqtSlot()
    def move_down(self):
        """ Move selected event down one row """
        self._move(+1)

    def _move(self, direction):
        self._testcase.event_list.move(self._selected_event, direction)
        self._ui.tableEvents.selectRow(self.get_selected_index()+direction)

    @pyqtSlot()
    def update_selected_event(self):
        """ Sets the properties table to the selected event """
        item_index = self.get_selected_index()
        if not self._recorder.is_recording:
            if item_index < 0:
                self._testcase.event_list.offset = self._testcase.event_list.size()
            else:
                self._testcase.event_list.offset = item_index+1

        if 0 <= item_index < self._testcase.event_list.size():
            self._selected_event = self._testcase.event_list.events[item_index]
        else:
            self._selected_event = None

        self._update_control_buttons(item_index)
        self._update_properties()

    def _update_control_buttons(self, index):
        """ Updates enableness of the buttons of the selected event """
        is_event_selected = self._selected_event is not None
        self._ui.eventMgmtButtons.ui.buttonDel.setEnabled(is_event_selected)
        self._ui.eventMgmtButtons.ui.buttonUp.setEnabled(is_event_selected and index > 0)
        self._ui.eventMgmtButtons.ui.buttonDown.setEnabled(is_event_selected and index < self._testcase.event_list.size()-1)

    def _update_properties(self):
        """ Updates the property view of the selected event """
        if self._selected_event is not None:
            self._ui.tableProperties.setEnabled(True)
            model = EventPropertiesModel(self._selected_event, self)
            self._ui.tableProperties.setModel(model)
        else:
            self._selected_event = None
            self._ui.tableProperties.setEnabled(False)
            self._ui.tableProperties.setModel(EventPropertiesModel(None, self))

    @pyqtSlot(str)
    def change_event_type(self, new_class):
        """ Change the type of the selected event (may be unsafe) """
        if self._selected_event is not None:
            event = Event.create(new_class, self._selected_event)
            self._testcase.event_list.replace(self._selected_event, event)
            self._selected_event = event
            self._update_properties()

    @pyqtSlot(tuple)
    def set_shortcut_modifiers(self, modifiers):
        if self._selected_event is not None:
            self._selected_event.modifiers = modifiers
            self._update_properties()


 ######  ##    ##    ###    ########   ######  ##     ##  #######  ########  ######
##    ## ###   ##   ## ##   ##     ## ##    ## ##     ## ##     ##    ##    ##    ##
##       ####  ##  ##   ##  ##     ## ##       ##     ## ##     ##    ##    ##
 ######  ## ## ## ##     ## ########   ######  ######### ##     ##    ##     ######
      ## ##  #### ######### ##              ## ##     ## ##     ##    ##          ##
##    ## ##   ### ##     ## ##        ##    ## ##     ## ##     ##    ##    ##    ##
 ######  ##    ## ##     ## ##         ######  ##     ##  #######     ##     ######


    def get_selected_snap_index(self):
        """
        :return: The index of the selected row of the snapshot list,
                 or -1 if the selection is empty
        """
        selected_rows = self._ui.listSnap.selectionModel().selectedRows()
        if selected_rows:
            return selected_rows[0].row()
        return -1


    @pyqtSlot()
    def move_snap_up(self):
        """ Move selected event up one row """
        self._move_snap(-1)

    @pyqtSlot()
    def move_snap_down(self):
        """ Move selected event down one row """
        self._move_snap(+1)

    def _move_snap(self, direction):
        self._testcase.snap_list.move(self._selected_snap, direction)
        self._ui.listSnap.setCurrentIndex(
            self._ui.listSnap.model().index(self.get_selected_snap_index()+direction, 0)
            )

    @pyqtSlot()
    def update_selected_snap(self):
        """
        Sets the properties table to the selected snapshot and updates the
        snap preview
        """
        item_index = self.get_selected_snap_index()
        if 0 <= item_index < self._testcase.snap_list.size():
            self._selected_snap = self._testcase.snap_list.get(item_index)
            img = self._selected_snap.as_image()
            if img is not None:
                self._ui.labelPreview.setPixmap(img)
        else:
            self._selected_snap = None
            # reset to default image
            self._ui.labelPreview.setPixmap(QPixmap(":/eyecandy/ui/icons/snitch_a30.png"))

        self._update_snap_properties()
        self._update_snap_control_buttons(item_index)

    def _update_snap_control_buttons(self, index):
        """ Updates enableness of the buttons of the selected snapshot """
        is_snap_selected = self._selected_snap is not None
        self._ui.capsEventMgmtButtons.ui.buttonDel.setEnabled(is_snap_selected)
        self._ui.capsEventMgmtButtons.ui.buttonUp.setEnabled(is_snap_selected and index > 0)
        self._ui.capsEventMgmtButtons.ui.buttonDown.setEnabled(is_snap_selected and index < self._testcase.snap_list.size()-1)

    def _update_snap_properties(self):
        """ Updates the property view of the selected snapshots """
        if self._selected_snap is not None:
            self._ui.tableSnapProps.setEnabled(True)
            model = PropertiesTableModel(self._selected_snap, self)
            self._ui.tableSnapProps.setModel(model)
        else:
            self._ui.tableSnapProps.setEnabled(False)
            self._ui.tableSnapProps.setModel(EventPropertiesModel(None, self))

    @pyqtSlot()
    def capture(self):
        """ Starts the screen capture sequence """
        QApplication.setOverrideCursor(Qt.CrossCursor)
        self.snap = Snapper()
        self.snap.area_captured.connect(self.save_capture)

    @pyqtSlot(int, int, int, int)
    def save_capture(self, x, y, w, h):
        """ Slot called when the capture is performed """
        self.snap = None

        capture = Screenshot(region=(x, y, w, h))
        capture.do_capture()
        self._testcase.snap_list.push(capture)
        self._ui.listSnap.setCurrentIndex(self._ui.listSnap.model().index(self._testcase.snap_list.size()-1, 0, QModelIndex()))
        QApplication.restoreOverrideCursor()

    @pyqtSlot()
    def diff(self):
        """ Shows the image diff window for the selected capture """
        selection = self._ui.listSnap.selectedIndexes()
        if selection:
            snap_id = selection[0].row()
            self._snap_player.show_diff(snap_id)
        else:
            QMessageBox.warning(
                self,
                self.tr("Error"),
                self.tr("Please select a snapshot.")
            )

    @pyqtSlot()
    def ocr(self):
        """ Computes the ocr of the selected capture """
        selection = self._ui.listSnap.selectedIndexes()
        if selection:
            snap_id = selection[0].row()
            self._snap_player.compute_ocr(snap_id)
            self._update_snap_properties()
        else:
            QMessageBox.warning(
                self,
                self.tr("Error"),
                self.tr("Please select a snapshot.")
            )


######## #### ##       ########    ##     ##  ######   ##     ## ########
##        ##  ##       ##          ###   ### ##    ##  ###   ###    ##
##        ##  ##       ##          #### #### ##        #### ####    ##
######    ##  ##       ######      ## ### ## ##   #### ## ### ##    ##
##        ##  ##       ##          ##     ## ##    ##  ##     ##    ##
##        ##  ##       ##          ##     ## ##    ##  ##     ##    ##    ###
##       #### ######## ########    ##     ##  ######   ##     ##    ##    ###


    def new_testcase(self):
        self._testcase = TestCase(self)

        self._recorder = EventRecorder(self._testcase.event_list, parent=self)
        self._player = EventPlayer(self._testcase.event_list, parent=self)
        self._ui.tableEvents.setModel(EventListModel(self._testcase.event_list, self))

        self._snap_player = SnapshotPlayer(self._testcase.snap_list, parent=self)
        self._ui.listSnap.setModel(SnapListModel(self._testcase.snap_list, self))

        self._ui.eventMgmtButtons.ui.buttonDel.clicked.connect(lambda: self._testcase.event_list.remove(self._selected_event))
        self._ui.capsEventMgmtButtons.ui.buttonDel.clicked.connect(lambda: self._testcase.snap_list.remove(self._selected_snap))

        self._ui.buttonRecord.clicked.connect(self._recorder.start)
        self._ui.buttonStop.clicked.connect(self._recorder.stop)
        self._ui.buttonClear.clicked.connect(self._clear)
        self._ui.buttonPlaySnaps.clicked.connect(self._snap_player.play)

        self._testcase.event_list.list_changed.connect(self.update_status)
        self._testcase.event_list.list_changed.connect(self.update_selected_event)
        self._testcase.snap_list.list_changed.connect(self.update_selected_snap)


    @pyqtSlot()
    def show_save_dialog(self):
        """ Opens the file save dialog """
        filename, filetype = QFileDialog.getSaveFileName(
            self,
            self.tr("Save as"),
            CFG['history']['last_location']+"/record.json",
            ';;'.join(Data.NAMES),
            dict(zip(Data.TYPES, Data.NAMES))[Data.DEFAULT_TYPE]
        )
        if filename:
            self._testcase.save(filename, dict(zip(Data.NAMES, Data.TYPES))[filetype])
            SETTINGS.append_to_history(filename)
            self.update_open_recent_menu()


    @pyqtSlot()
    def show_open_dialog(self):
        """ Opens the file open dialog """
        filename, filetype = QFileDialog.getOpenFileName(
            self,
            self.tr("Open"),
            CFG['history']['last_location'],
            ';;'.join(Data.NAMES),
            dict(zip(Data.TYPES, Data.NAMES))[Data.DEFAULT_TYPE]
        )
        if filename:
            self._testcase.event_list.clear()
            self._testcase.snap_list.clear()
            self._testcase.load(filename, dict(zip(Data.NAMES, Data.TYPES))[filetype])
            SETTINGS.append_to_history(filename)
            self.update_open_recent_menu()

    def load(self, filename):
        if os.path.isfile(filename):
            self._testcase.load(filename)
            return True
        return False

    def save_result(self, filename):
        self._testcase.save(filename.replace('.json', '.result.json'))

    def update_open_recent_menu(self):
        self._ui.menuOpenRecent.clear()
        for recent in CFG['history']['last_files']:
            self._ui.menuOpenRecent.addAction(recent)

    @pyqtSlot(QAction)
    def open_recent(self, action):
        filename = action.text()
        if os.path.isfile(filename):
            self._testcase.load(filename)
            SETTINGS.append_to_history(action.text())
        else:
            QMessageBox.information(
                self,
                self.tr("Information"),
                self.tr("File doesn’t exist anymore.")
            )
            CFG['history']['last_files'].remove(filename)

        self.update_open_recent_menu()


##     ## ####  ######   ######         ##     ## ####
###   ###  ##  ##    ## ##    ##        ##     ##  ##
#### ####  ##  ##       ##              ##     ##  ##
## ### ##  ##   ######  ##              ##     ##  ##
##     ##  ##        ## ##              ##     ##  ##
##     ##  ##  ##    ## ##    ## ###    ##     ##  ##
##     ## ####  ######   ######  ###     #######  ####


    @pyqtSlot()
    def update_status(self):
        """ Updates the statusbar text on event list change """
        message = self.tr('Event list cleared.')
        if self._testcase.event_list.size() > 0:
            message = self.tr('Event added: {0}'.format(self._testcase.event_list.events[-1]))
        self._ui.statusbar.showMessage(message, 1000)
        self.update_status_label()

    @pyqtSlot()
    def update_status_label(self):
        """ Updates the default string of the statusbar """
        action = ''
        if self._recorder.is_recording:
            action = self.tr('RECORDING... ')
        self._status_label = action + Controller.STATUS_TEXT.format(self._testcase.event_list.size())
        self._ui.statusbar.showMessage(self._status_label)

    @pyqtSlot()
    def update_mouse_pos(self):
        self._ui.statuspos.setText('x={} ; y={}'.format(*pyautogui.position()))

    @pyqtSlot()
    def show_settings_dialog(self):
        """ Opens the settings and configuration dialog """
        SettingsDialog(self).show()

    @pyqtSlot()
    def show_about_dialog(self):
        """ Opens the information dialog """
        AboutDialog(self).show()

    @pyqtSlot()
    def _clear(self):
        if QMessageBox.question(
                self,
                self.tr("Confirmation"),
                self.tr("Are you sure you want to delete all events and screenshots?")
                ) == QMessageBox.Yes:

            self._testcase.event_list.clear()
            self._testcase.snap_list.clear()

    def closeEvent(self, event):
        # pylint: disable=no-self-use
        SETTINGS.save()
        event.accept()



"""
Module managing the input events generated by the user on the graphical interface.
"""
import logging as log
import time
import sys
import threading

import pyautogui
from pynput.keyboard import Key, KeyCode, Controller

def get_subclasses(base_class):
    """ :return: A list of all the classes inheriting, directly or not from the base_class. """
    subclasses = [base_class]
    for c in base_class.__subclasses__():
        subclasses += get_subclasses(c)
    return subclasses

def format_modifiers(modifier_names):
    """ :type modifier_names: list of names to join """
    return '{}+'.format('+'.join(modifier_names)) if modifier_names else ''

def use_modifiers(executor):
    """
    Decorator handling the press of key modifiers while executing other
    types of events.
    It is meant to be used on the `execute` method of Event sub-classes.
    """
    def wrapper(self):
        self.ctrl = Controller()
        modifier_keys = tuple(Key[k] for k in self.modifiers)
        if self.modifiers == (Key.alt_gr.name,):
            if sys.platform in ALT_GR:
                modifier_keys = ALT_GR[sys.platform]
            else:
                log.warning('Unrecognized platform "%s"', sys.platform)

        for m in modifier_keys:
            self.ctrl.press(m)
            log.debug('»»»» %s pressed', m)

        executor(self)

        for m in modifier_keys:
            self.ctrl.release(m)
            log.debug('»»»» %s released', m)

        del self.ctrl
    return wrapper

class Event:
    """
    Base class for events. Only contains the time of occurrence (in ms).
    All event classes must derive from this class and -at least- reimplement
    the __str___ and execute methods.
    """
    DOUBLE_CLICK_THRESHOLD = 300
    MOUSE_MOVE_DURATION = .5
    MOUSE_MOVE_EASING = pyautogui.easeInQuad

    @staticmethod
    def create(class_name, from_event=None):
        """
        Creates a new event of the specified class, while preserving the values
        of the origin event (as much as possible).
        :param class_name: The name of the type of event to create
        :param from_event: The event from wich the common values will be copied
        """
        #pylint: disable=eval-used
        new_event = eval(class_name)()
        if from_event is not None:
            for key in new_event.__dict__.keys():
                # collecting old values while preserving new type
                if key in from_event.__dict__ and key != 'type':
                    new_event.__dict__[key] = from_event.__dict__[key]
        return new_event


    @staticmethod
    def get_subclasses():
        """ :return: A list of all the classes inheriting, directly or not from this class. """
        subclasses = get_subclasses(Event)
        subclasses.remove(Event)
        return subclasses

    def __init__(self):
        self.time = round(time.time()*1000)
        self.type = self.__class__.__name__
        self.block_events_on_exec = True
        log.debug(self)

    def __str__(self):
        return 'EVT: @{0}'.format(self.time)

    def execute(self):
        """ Replays the event """
        log.debug(str(self))

    def get_serializable(self):
        return self.__dict__


##     ##  #######  ##     ##  ######  ########    ######## ##     ## ######## ##    ## ########  ######
###   ### ##     ## ##     ## ##    ## ##          ##       ##     ## ##       ###   ##    ##    ##    ##
#### #### ##     ## ##     ## ##       ##          ##       ##     ## ##       ####  ##    ##    ##
## ### ## ##     ## ##     ##  ######  ######      ######   ##     ## ######   ## ## ##    ##     ######
##     ## ##     ## ##     ##       ## ##          ##        ##   ##  ##       ##  ####    ##          ##
##     ## ##     ## ##     ## ##    ## ##          ##         ## ##   ##       ##   ###    ##    ##    ##
##     ##  #######   #######   ######  ########    ########    ###    ######## ##    ##    ##     ######


class MouseImageClick(Event):
    """
    Class representing a single click with the main mouse button on a location
    specified by an image.
    """
    def __init__(self, image_path='', image_index=0):
        """
        :param image_path: The location of the capture of the area to clik on
        :param image_index: If multiple area on the screen exactly match the
                            capture, this is the 0-based index of the area to
                            click on. The numbering starts on the top-left
                            corner of the screen.
        """
        self.path = image_path
        self.index = image_index
        super().__init__()

    def __str__(self):
        return 'CLICK: [{0}#{1}]'.format(self.path, self.index)

    def execute(self):
        locations = pyautogui.locateAllOnScreen(self.path)
        if locations and len(locations) > self.index:
            pyautogui.click(pyautogui.center(locations[self.index]))
        else:
            log.warning('The area defined by "%s#%d" couldn\'t be found',
                        self.path, self.index)


class MouseClick(Event):
    """ Class representing a single click with the main mouse button """
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y
        self.modifiers = []
        super().__init__()

    def __str__(self):
        return '{2}CLICK: [x={0}, y={1}]'.format(
            self.x,
            self.y,
            format_modifiers(self.modifiers))

    @use_modifiers
    def execute(self):
        log.debug('»»» moving to (%d, %d)', self.x, self.y)
        pyautogui.moveTo(self.x, self.y, Event.MOUSE_MOVE_DURATION, Event.MOUSE_MOVE_EASING)
        log.debug('»»» click!')
        # down then up instead of pyautogui.click() to prevent Java missing event
        pyautogui.mouseDown()
        pyautogui.mouseUp()


class MouseDrag(MouseClick):
    """ Class representing a click-drag-release with the mouse """
    def __init__(self, x0=0, y0=0, x1=0, y1=0):
        """
        :parm start_location: The coordinates where the mouse clicked
        :type start_location: A two elements tuple
        :parm end_location: The coordinates where the mouse dropped the dragged element
        :type end_location: A two elements tuple
        """
        self.xx = x1
        self.yy = y1
        super().__init__(x0, y0)

    def __str__(self):
        return 'DRAG: [x0={0}, y0={1}] -> [x1={2}, y1={3}]'.format(
            self.x, self.y, self.xx, self.yy)

    @use_modifiers
    def execute(self):
        log.debug('»»» dragging from (%d,%d) to (%d,%d)', self.x, self.y, self.xx, self.yy)
        log.debug('»»»   moving to (%d, %d)', self.x, self.y)
        pyautogui.moveTo(self.x, self.y, Event.MOUSE_MOVE_DURATION)
        log.debug('»»»   pressing mouse button')
        pyautogui.mouseDown()
        log.debug('»»»   moving to (%d, %d)', self.xx, self.yy)
        pyautogui.moveTo(self.xx, self.yy, Event.MOUSE_MOVE_DURATION)
        log.debug('»»»   releasing mouse button')
        pyautogui.mouseUp()
        log.debug('»»» .')


class MouseAltClick(MouseClick):
    """ Class representing a single click with the secondary mouse button """

    def __str__(self):
        return '{2}ALT CLICK: [x={0}, y={1}]'.format(
            self.x,
            self.y,
            format_modifiers(self.modifiers))

    @use_modifiers
    def execute(self):
        log.debug('»»» moving to (%d, %d)', self.x, self.y)
        pyautogui.moveTo(self.x, self.y, Event.MOUSE_MOVE_DURATION, Event.MOUSE_MOVE_EASING)
        log.debug('»»» click (alt)')
        pyautogui.rightClick()
        log.debug('»»» .')


class MouseDoubleClick(MouseClick):
    """ Class representing a doube click with the main mouse button """

    def __str__(self):
        return '{2}DOUBLE CLICK: [x={0}, y={1}]'.format(
            self.x,
            self.y,
            format_modifiers(self.modifiers))

    def execute(self):
        log.debug('»»» moving to (%d, %d)', self.x, self.y)
        pyautogui.moveTo(self.x, self.y, Event.MOUSE_MOVE_DURATION, Event.MOUSE_MOVE_EASING)
        log.debug('»»» click, click!')
        pyautogui.doubleClick()
        log.debug('»»» .')


##    ## ######## ##    ##    ######## ##     ## ######## ##    ## ########  ######
##   ##  ##        ##  ##     ##       ##     ## ##       ###   ##    ##    ##    ##
##  ##   ##         ####      ##       ##     ## ##       ####  ##    ##    ##
#####    ######      ##       ######   ##     ## ######   ## ## ##    ##     ######
##  ##   ##          ##       ##        ##   ##  ##       ##  ####    ##          ##
##   ##  ##          ##       ##         ## ##   ##       ##   ###    ##    ##    ##
##    ## ########    ##       ########    ###    ######## ##    ##    ##     ######


MODIFIERS = (Key.ctrl, Key.ctrl_l, Key.ctrl_r,
             Key.alt, Key.alt_l, Key.alt_r, Key.alt_gr,
             Key.shift, Key.shift_l, Key.shift_r,
             KeyCode(65027))

class TextEntered(Event):
    """ Class representing a sequence of key strokes on the keyboard """
    def __init__(self, text=' '):
        """
        :parm text: The coordinates where the mouse clicked
        """
        self.text = str(text.char) if isinstance(text, KeyCode) else str(text)
        super().__init__()

    def __str__(self):
        txt = ''
        if self.text is not None:
            txt = ''.join(self.text)
        return 'TYPE: [txt={0}]'.format(txt)

    def execute(self):
        # DON'T USE PYAUTOGUI FOR THIS
        kbd = Controller()
        kbd.type(self.text)
        log.debug('»»» %s', self.text)


VK_TO_KEY = dict((k.value.vk, k.name) for k in Key)

class KeyPressed(Event):
    def __init__(self, key=Key.esc):
        """ :type key: Key or KeyCode or the name as a string """
        if isinstance(key, Key):
            self.key = key.name
        elif isinstance(key, KeyCode):
            log.warning('KeyPressed registered a KeyCode: %s', key)
            if key.char is not None:
                self.key = key.char
            elif key.vk in VK_TO_KEY:
                self.key = VK_TO_KEY[key.vk]
            else:
                log.warning('Not a known KeyCode: %s', key)
                self.key = key.vk
        elif isinstance(key, str):
            if key in Key.__members__:
                self.key = key
            else:
                log.error('Not a valid key name: %s', key)
                raise TypeError()
        else:
            log.error('Wrong type of key: %s', key)
            raise TypeError()
        super().__init__()

    def __str__(self):
        return 'KEY: [{0}]'.format(self.key)

    def execute(self):
        ctrl = Controller()
        if isinstance(self.key, str):
            k = Key[self.key]
        elif isinstance(self.key, int):
            k = KeyCode(self.key)
        else:
            raise TypeError()

        ctrl.press(k)
        log.debug('»»» %s pressed', self.key)
        ctrl.release(k)
        log.debug('»»» %s released', self.key)

ALT_GR = {
    'win32': (Key.ctrl_l, Key.alt_r),
    'linux': (KeyCode(65027),)
}

class Shortcut(KeyPressed):
    def __init__(self, key=Key.esc, modifiers=None):
        self.modifiers = tuple()
        if modifiers is not None:
            if tuple(modifiers) in ALT_GR.values():
                self.modifiers = (Key.alt_gr.name,)
            else:
                self.modifiers = tuple(k.name for k in modifiers)
        super().__init__(key)

    def __str__(self):
        return 'SHORTCUT: [{0}+{1}]'.format('+'.join(self.modifiers), self.key)

    @use_modifiers
    def execute(self):
        log.debug('»»» %s', self)
        if self.key in Key.__members__:
            k = Key[self.key]
        else:
            k = self.key

        # pylint: disable=no-member
        ctrl = self.ctrl if hasattr(self, 'ctrl') else Controller()

        ctrl.press(k)
        log.debug('»»»» %s pressed', k)
        ctrl.release(k)
        log.debug('»»»» %s released', k)


 #######  ######## ##     ## ######## ########     ######## ##     ## ######## ##    ## ########  ######
##     ##    ##    ##     ## ##       ##     ##    ##       ##     ## ##       ###   ##    ##    ##    ##
##     ##    ##    ##     ## ##       ##     ##    ##       ##     ## ##       ####  ##    ##    ##
##     ##    ##    ######### ######   ########     ######   ##     ## ######   ## ## ##    ##     ######
##     ##    ##    ##     ## ##       ##   ##      ##        ##   ##  ##       ##  ####    ##          ##
##     ##    ##    ##     ## ##       ##    ##     ##         ## ##   ##       ##   ###    ##    ##    ##
 #######     ##    ##     ## ######## ##     ##    ########    ###    ######## ##    ##    ##     ######


class Delay(Event):
    """ Class representing pause in the action (waiting for something to complete) """
    def __init__(self, duration=.5):
        """
        :type duration: int
        :param duration: Duration of the waiting time, in milliseconds
        """
        self.duration = duration
        self.block_events_on_exec = False
        self.exit = threading.Event()
        super().__init__()

    def __str__(self):
        return 'PAUSE: [dur={0}s]'.format(self.duration)

    def execute(self):
        log.debug('»»» waiting %fs', self.duration)
        self.exit.wait(self.duration)
        log.debug('»»» .')

    def kill(self):
        self.exit.set()

    def get_serializable(self):
        items = self.__dict__.copy()
        del items['exit']
        del items['block_events_on_exec']
        return items

class Comment(Event):
    """ Class for comment event (just does nothing on execution) """
    def __init__(self, text=''):
        self.text = text
        super().__init__()

    def __str__(self):
        return '/*** {} ***/'.format(self.text)

    def execute(self):
        log.debug('»»» /*** %s ***/', self.text)


class EventSequence(Event):
    """ A list of events"""
    DEFAULT_DELAY = 200

    def __init__(self, event_list=None):
        super().__init__()

        self.events = []
        if event_list is not None:
            self.events = event_list
        self.delay = EventSequence.DEFAULT_DELAY

    def __str__(self):
        return "{0} events in list".format(len(self.events))

    def execute(self):
        for e in self.events:
            e.execute()



"""
Module providing tools for recordin user inputs (mouse clicks and keyboard strokes)
"""
import time
import logging as log
import traceback
import sys
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

import pyautogui
from pynput.keyboard import Key, KeyCode
from pynput.keyboard import Listener as KbdListener
from pynput.mouse import Button
from pynput.mouse import Listener as MouseListener

from .model.data.events import KeyPressed, TextEntered, Shortcut, MouseAltClick, MouseClick, MouseDrag, MODIFIERS
from .settings import CFG

#pylint: disable=unused-argument

def print_err():
    """ Prints error message after exception """
    t, message, trace = sys.exc_info()
    log.error("%s: %s", t.__name__, message)
    log.error("\n".join(traceback.format_tb(trace)))

class SnapInfo:
    """
    Class containing informations about various sizes used to
    perform screen captures.
    """

    CAPTURE_SIZE = 100
    SCREEN_WIDTH, SCREEN_HEIGHT = pyautogui.size()

    @staticmethod
    def get_area(x, y):
        """
        :return: A 4-item tuple containing the left, top, width, height of a square
                 area of CAPTURE_SIZ side length and centered in (x, y). If the click
                 is too close from the border of the screen, the area captured is
                 preserved, but the click is not centered
        """
        return (max(0, min(x-SnapInfo.CAPTURE_SIZE, SnapInfo.SCREEN_WIDTH)),
                max(0, min(y-SnapInfo.CAPTURE_SIZE, SnapInfo.SCREEN_HEIGHT)),
                SnapInfo.CAPTURE_SIZE,
                SnapInfo.CAPTURE_SIZE)


class KeyEventCatcher:
    """
    A one-shot key combination catcher.
    Starts to listen to key events on call to method start
    or on creation if used through the `with` keyword.
    """

    def __init__(self, keys, callback):
        """
        :type keys: list of set of pressed keys
        :type callback: callable to call when keys are pressed
        """

        self._shortcut = keys
        self._callback = callback
        self._pressed_keys = set()
        self._kl = None
        self._is_running = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def start(self):
        log.debug('KeyEventCatcher START')
        self._kl = KbdListener(on_press=self.on_press,
                               on_release=self.on_release)
        self._kl.start()
        self._is_running = True

    def stop(self):
        log.debug('KeyEventCatcher STOP')
        self._kl.stop()
        self._is_running = False

    def on_press(self, key):
        """ Callback for keyboard press events """
        if not self._is_running:
            log.info('PressEvent occured on stopped state, killing listener.')
            raise KbdListener.StopException()
        try:
            log.debug('!vvv %s', key)
            self._pressed_keys.add(key)
            log.debug('    : %s',
                     list(map(lambda l: l.name if isinstance(l, Key) else l.char, self._pressed_keys))
                     )

            if self._pressed_keys in self._shortcut:
                log.debug('Shortcut detected, triggering callback')
                self.stop()
                self._callback()
        #pylint: disable=bare-except
        except:
            print_err()

    def on_release(self, key):
        if not self._is_running:
            log.info('ReleaseEvent occured on stopped state, killing listener.')
            raise KbdListener.StopException()
        log.debug('!^^^ %s (%s)', key, type(key))
        if key in self._pressed_keys:
            self._pressed_keys.remove(key)



class EventRecorder(QObject):
    recording_started = pyqtSignal()
    recording_stopped = pyqtSignal()
    recording_paused = pyqtSignal()

    """ Class recording user input event into an EventList """
    def __init__(self, event_list, snap_click=False, parent=None):
        """
        :type event_list: EventList
        :param snap_click: If true, each mouse click generaetes a small screen capture
                           at the click location.
        """
        super().__init__(parent)

        self.event_list = event_list
        self.is_recording = False
        self._init_listeners()
        self._snap_on_click = snap_click
        self._pressed_keys = set()
        self._press_position = (0, 0)
        self._ignore_next = set()

    def _init_listeners(self):
        """ Creates new listener objects for mouse and keyboard """
        self._ml = MouseListener(on_click=self.on_click)
        self._kl = KbdListener(on_press=self.on_press,
                               on_release=self.on_release)


##     ##  #######  ##     ##  ######  ########
###   ### ##     ## ##     ## ##    ## ##
#### #### ##     ## ##     ## ##       ##
## ### ## ##     ## ##     ##  ######  ######
##     ## ##     ## ##     ##       ## ##
##     ## ##     ## ##     ## ##    ## ##
##     ##  #######   #######   ######  ########


    def on_click(self, x, y, button, pressed):
        """ Callback for click events """
        x, y = round(x), round(y)
        try:
            pressed_modifiers = {m for m in MODIFIERS if m in self._pressed_keys}
            if self._snap_on_click:
                image = pyautogui.screenshot(region=SnapInfo.get_area(x, y))
                image.save('{0}-click-at-{1}x{2}.png'.format(round(time.time()*1000), x, y), 'PNG')
            event = None
            if pressed:
                self._press_position = (x, y)
                if button == Button.left:
                    event = MouseClick(x, y)
                else: # Button.right
                    event = MouseAltClick(x, y)
            else:
                if button == Button.left and self._press_position != (x, y):
                    press = self.event_list.pop()
                    event = MouseDrag(x0=press.x, y0=press.y, x1=x, y1=y)
                #else: pass, this is a regular click

            if event:
                event.modifiers = tuple(m.name for m in pressed_modifiers)
                self.event_list.push(event)

        #pylint: disable=bare-except
        except:
            print_err()


##    ## ######## ##    ## ########   #######     ###    ########  ########
##   ##  ##        ##  ##  ##     ## ##     ##   ## ##   ##     ## ##     ##
##  ##   ##         ####   ##     ## ##     ##  ##   ##  ##     ## ##     ##
#####    ######      ##    ########  ##     ## ##     ## ########  ##     ##
##  ##   ##          ##    ##     ## ##     ## ######### ##   ##   ##     ##
##   ##  ##          ##    ##     ## ##     ## ##     ## ##    ##  ##     ##
##    ## ########    ##    ########   #######  ##     ## ##     ## ########


    def on_press(self, key):
        """ Callback for keyboard press events """
        try:
            self._pressed_keys.add(key)
            log.debug('vvv %s (%s): %s', key, type(key).__name__, self._get_pressed_names())

            modifiers = {k for k in self._pressed_keys if k in MODIFIERS}

            if not key in MODIFIERS:
                if modifiers:
                    self.event_list.push(Shortcut(key, modifiers))
                else:
                    if isinstance(key, Key):
                        self.event_list.push(KeyPressed(key))
                    elif isinstance(key, KeyCode):
                        if key.char:
                            self.event_list.push(TextEntered(key.char))
                        else:
                            self.event_list.push(KeyPressed(key))
                    else:
                        self.event_list.push(TextEntered(key))
        #pylint: disable=bare-except
        except:
            print_err()

    def on_release(self, key):
        """ Callback for keyboard release events """
        try:
            if key in self._pressed_keys:
                self._pressed_keys.remove(key)
            log.debug('^^^ %s: %s', key, self._get_pressed_names())

        #pylint: disable=bare-except
        except:
            print_err()

    def _get_pressed_names(self):
        return list(map(lambda l: l.name if isinstance(l, Key) else l.char, self._pressed_keys))

########  ########  ######   #######  ########  ########  #### ##    ##  ######
##     ## ##       ##    ## ##     ## ##     ## ##     ##  ##  ###   ## ##    ##
##     ## ##       ##       ##     ## ##     ## ##     ##  ##  ####  ## ##
########  ######   ##       ##     ## ########  ##     ##  ##  ## ## ## ##   ####
##   ##   ##       ##       ##     ## ##   ##   ##     ##  ##  ##  #### ##    ##
##    ##  ##       ##    ## ##     ## ##    ##  ##     ##  ##  ##   ### ##    ##
##     ## ########  ######   #######  ##     ## ########  #### ##    ##  ######


    @pyqtSlot()
    def start(self):
        """ Starts the event recording """
        self.is_recording = True
        self._ml.start()
        self._kl.start()

        if CFG['minimize_on_record']:
            self.parent().setWindowState(self.parent().windowState() | Qt.WindowMinimized)
        self.recording_started.emit()

    @pyqtSlot()
    def pause(self):
        """ Pauses the event recording """
        self.is_recording = False
        self._ml.stop()
        self._kl.stop()
        self.event_list.pop()
        self.recording_paused.emit()

    @pyqtSlot()
    def stop(self):
        """ Stops the event recording """
        self.is_recording = False
        self._ml.stop()
        self._kl.stop()
        self._init_listeners()
        self.event_list.pop()
        if CFG['minimize_on_record']:
            self.parent().setWindowState(self.parent().windowState() & ~Qt.WindowMinimized | Qt.WindowActive)

        self._ignore_next.clear()
        self._pressed_keys.clear()

        self.recording_stopped.emit()

    def reset(self):
        """ Clears the event recording """
        self.event_list.clear()



"""
Package containing classes derived from QObjects defining models handling events
"""
import logging as log

from PyQt5.QtCore import Qt
from PyQt5.QtCore import QAbstractTableModel, QVariant, QModelIndex


class ListModel(QAbstractTableModel):
    """ Base class for TableModels handling lists """

    def __init__(self, data, parent=None):
        """
        :type data: EventList
        """
        super().__init__(parent)

        self._data = data
        self._last_size = data.size()

        self._data.list_changed.connect(self.update_list)
        self._last_col = 0

    #region## QAbstractTableModel subclassing
    def rowCount(self, parent):
        #pylint: disable=unused-argument
        """ Gives the number of events in the list """
        return self._data.size()

    def columnCount(self, parent):
        #pylint: disable=unused-argument, no-self-use
        return 1

    def data(self, index, role=Qt.DisplayRole):
        #pylint: disable=unused-argument, no-self-use
        return QVariant()

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        #pylint: disable=unused-argument, no-self-use
        return QVariant()
    #endregion##

    def update_list(self):
        """ Handles data list change """
        size = self._data.size()
        last_row = size-1

        if size > self._last_size:
            log.debug('Inserting %d -> %d', self._last_size, size)
            self.beginInsertRows(QModelIndex(), self._last_size, last_row)
            self.endInsertRows()
        elif size < self._last_size:
            log.debug('Removing %d -> %d', self._last_size, size)
            self.beginRemoveRows(QModelIndex(), last_row+1, self._last_size-1)
            self.endRemoveRows()

        self.dataChanged.emit(self.index(0, 0), self.index(last_row, self._last_col))
        self._last_size = size



import time
import logging as log

from PyQt5.QtGui import QColor, QPainter, QBrush
from PyQt5.QtWidgets import QWidget
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, QPoint, QRect, pyqtSignal

class Snapper(QWidget):
    area_captured = pyqtSignal(int, int, int, int)

    def __init__(self):
        super().__init__(None, Qt.NoDropShadowWindowHint | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)

        self.begin = QPoint()
        self.end = QPoint()

        self.setAttribute(Qt.WA_NoSystemBackground, True)
        self.setAttribute(Qt.WA_TranslucentBackground, True)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.widget = QtWidgets.QLabel('', self)
        self.layout.addWidget(self.widget)
        self.widget.setStyleSheet('background: rgba(0, 0, 0, .1);')
        self.setGeometry(30, 30, 800, 450)
        self.showFullScreen()


    def _reset(self):
        self.begin = QPoint()
        self.end = QPoint()

    # pylint: disable=unused-argument
    def paintEvent(self, event):
        qp = QPainter(self)
        br = QBrush(QColor(255, 192, 192, 60))
        qp.setBrush(br)
        qp.drawRect(QRect(self.begin, self.end))

    def mousePressEvent(self, event):
        self.begin = event.pos()
        self.end = event.pos()
        self.update()

    def mouseMoveEvent(self, event):
        self.end = event.pos()
        self.update()

    def mouseReleaseEvent(self, event):
        self.end = event.pos()
        self.close()

        x = min(self.begin.x(), self.end.x())
        y = min(self.begin.y(), self.end.y())
        w = abs(self.begin.x() - self.end.x())
        h = abs(self.begin.y() - self.end.y())

        time.sleep(.4) # prevents rectangle selection to be captured
        log.info('Area captured X=%d Y=%d %dx%d', x, y, w, h)
        self.area_captured.emit(x, y, w, h)



import logging as log

from PyQt5.QtCore import Qt, QCoreApplication
from PyQt5.QtCore import QAbstractTableModel, QVariant
from PyQt5.QtGui import QFontDatabase


class PropertiesTableModel(QAbstractTableModel):
    HEADERS = [
        QCoreApplication.translate('EventListModel', 'Name'),
        QCoreApplication.translate('EventListModel', 'Value')
    ]


    def __init__(self, element, parent=None):
        """
        :param element: Object of which the properties (__dict__) will be listed
        """
        super().__init__(parent)

        self._dict = None
        if element is not None:
            self._dict = element.__dict__

    def set_item(self, item):
        """
        :type item: Any subclass of element
        """
        self._dict = item.__dict__

    #region## QAbstractTableModel subclassing
    def rowCount(self, parent):
        #pylint: disable=unused-argument
        """ Gives the number of events in the list """
        if self._dict is not None:
            return len(self._dict)
        return 0

    def columnCount(self, parent):
        #pylint: disable=unused-argument, no-self-use
        """ There's 2 columns in the table: Timestamp, Type and Description string"""
        return 2

    def data(self, index, role=Qt.DisplayRole):
        """
        Returns the data for the specified cell for the specified role
        :param index: The table cell
        :type index: QModelIndex
        :param role: The Qt role of the returned value (only DisplayRole is implemented)
        """
        row = index.row()
        col = index.column()
        value = QVariant()
        if role in [Qt.DisplayRole, Qt.EditRole]:
            if col == 0:
                value = list(self._dict.keys())[row]
            elif col == 1:
                value = list(self._dict.values())[row]
        elif role == Qt.FontRole:
            font = QFontDatabase.systemFont(QFontDatabase.GeneralFont) #super().data(index, role)
            font.setPointSize(8)
            if col == 0:
                font.setBold(True)
            value = font

        return value

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        #pylint: disable=no-self-use
        """ The table column titles """
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return PropertiesTableModel.HEADERS[section]
            return section
        return QVariant()

    ### Edition
    def flags(self, index):
        """ Defines cell properties """
        flags = super().flags(index)
        if index.column() == 1:
            return flags | Qt.ItemIsEditable
        return flags

    def setData(self, index, value, role=Qt.EditRole):
        """ Sets data to model """
        #pylint: disable=unused-argument
        key = self.data(self.index(index.row(), 0))
        try:
            data_type = type(self._dict[key])
            self._dict[key] = data_type(value)
            self.dataChanged.emit(index, index)
            return True
        except TypeError:
            log.error('Data entered "%s" of type "%s" cannot be converted to "%s"',
                      value, type(value), data_type)
            return False
    #endregion##



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './snitch/ui/modifier_picker.ui'
#
# Created by: PyQt5 UI code generator 5.14.1
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ModifierCheckboxes(object):
    def setupUi(self, ModifierCheckboxes):
        ModifierCheckboxes.setObjectName("ModifierCheckboxes")
        ModifierCheckboxes.resize(190, 20)
        ModifierCheckboxes.setAutoFillBackground(True)
        self.horizontalLayout = QtWidgets.QHBoxLayout(ModifierCheckboxes)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.ctrlCheckBox = QtWidgets.QCheckBox(ModifierCheckboxes)
        font = QtGui.QFont()
        font.setPointSize(7)
        self.ctrlCheckBox.setFont(font)
        self.ctrlCheckBox.setFocusPolicy(QtCore.Qt.NoFocus)
        self.ctrlCheckBox.setIconSize(QtCore.QSize(16, 16))
        self.ctrlCheckBox.setObjectName("ctrlCheckBox")
        self.horizontalLayout.addWidget(self.ctrlCheckBox)
        self.shiftCheckBox = QtWidgets.QCheckBox(ModifierCheckboxes)
        font = QtGui.QFont()
        font.setPointSize(7)
        self.shiftCheckBox.setFont(font)
        self.shiftCheckBox.setFocusPolicy(QtCore.Qt.NoFocus)
        self.shiftCheckBox.setIconSize(QtCore.QSize(16, 16))
        self.shiftCheckBox.setObjectName("shiftCheckBox")
        self.horizontalLayout.addWidget(self.shiftCheckBox)
        self.altCheckBox = QtWidgets.QCheckBox(ModifierCheckboxes)
        font = QtGui.QFont()
        font.setPointSize(7)
        self.altCheckBox.setFont(font)
        self.altCheckBox.setFocusPolicy(QtCore.Qt.NoFocus)
        self.altCheckBox.setIconSize(QtCore.QSize(16, 16))
        self.altCheckBox.setObjectName("altCheckBox")
        self.horizontalLayout.addWidget(self.altCheckBox)
        self.alt_grCheckBox = QtWidgets.QCheckBox(ModifierCheckboxes)
        font = QtGui.QFont()
        font.setPointSize(7)
        self.alt_grCheckBox.setFont(font)
        self.alt_grCheckBox.setFocusPolicy(QtCore.Qt.NoFocus)
        self.alt_grCheckBox.setIconSize(QtCore.QSize(16, 16))
        self.alt_grCheckBox.setObjectName("alt_grCheckBox")
        self.horizontalLayout.addWidget(self.alt_grCheckBox)

        self.retranslateUi(ModifierCheckboxes)
        QtCore.QMetaObject.connectSlotsByName(ModifierCheckboxes)

    def retranslateUi(self, ModifierCheckboxes):
        _translate = QtCore.QCoreApplication.translate
        ModifierCheckboxes.setWindowTitle(_translate("ModifierCheckboxes", "Form"))
        self.ctrlCheckBox.setText(_translate("ModifierCheckboxes", "ctrl"))
        self.shiftCheckBox.setText(_translate("ModifierCheckboxes", "shift"))
        self.altCheckBox.setText(_translate("ModifierCheckboxes", "alt"))
        self.alt_grCheckBox.setText(_translate("ModifierCheckboxes", "alt gr"))



from PyQt5.QtWidgets import QWidget

from .list_manager_ui import Ui_EventMgmtButtons

class EventMgmtButtons(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.ui = Ui_EventMgmtButtons()
        self.ui.setupUi(self)



""" Snitch, keylogger and more """

__version__ = "0.6.0"
__author__ = "Grégory Millasseau"

from tempfile import gettempdir

LOG_LEVELS = ['CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']
DEFAULT_LOG_LEVEL = 'DEBUG'
DEFAULT_LOG_FILE = '{}/snitch.log'.format(gettempdir())
FILE_VERSION = "0.4"



from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot
from PyQt5.QtCore import QVariant

from .list import ListModel
from .data.screenshot import Screenshot


class SnapList(QObject):
    """
    Class managing a collection of snapshots.
    All items in the list of events must subclass Screenshot
    """
    list_changed = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.snaps = []

    def push(self, item):
        if isinstance(item, Screenshot):
            self.snaps.append(item)
            self.list_changed.emit()

    def get(self, index):
        if index < len(self.snaps):
            return self.snaps[index]
        return None

    def size(self):
        return len(self.snaps)

    def get_serializable(self):
        """ :return: this object as a dict """
        return {
            'screenshots': list(map(lambda l: l.__dict__, self.snaps))
        }

    @pyqtSlot(Screenshot)
    def remove(self, snap):
        if snap in self.snaps:
            self.snaps.remove(snap)
            self.list_changed.emit()
            return True
        return False

    def move(self, snap, direction):
        """
        Change snap position in list
        :param direction: +1 to move down, -1 to move up
        """
        if snap in self.snaps:
            index = self.snaps.index(snap)
            self.snaps[index+direction], self.snaps[index] = self.snaps[index], self.snaps[index+direction]
            self.list_changed.emit()
            return True
        return False

    @pyqtSlot()
    def clear(self):
        """ Removes all the elements from the list """
        del self.snaps[:]
        self.list_changed.emit()


class SnapListModel(ListModel):
    """ Model for displaying an SnapList in a QListView """

    def data(self, index, role=Qt.DisplayRole):
        """
        Returns the data for the specified cell for the specified role
        :param index: The table cell
        :type index: QModelIndex
        :param role: The Qt role of the returned value (only DisplayRole is implemented)
        """
        if role == Qt.DisplayRole:
            item = self._data.get(index.row())
            return str('Snapshot: x={0} y={1} size={2}x{3}'.format(item.x, item.y, item.w, item.h))

        return QVariant()



import os
import json
import logging as log

from PyQt5.QtCore import QObject, pyqtSlot

CONFIG_FILE = os.path.expanduser('~/.snitch.json')
DEFAULT_SETTINGS = {
  'minimize_on_record': False,
  'replay_delay': 0.2,
  'diff_threshold': 98,
  'history': {
    'max_size': 10,
    'last_location': '',
    'last_files': []
  }
}

class Settings(QObject):
    def __init__(self, load_file=True, parent=None):
        super().__init__(parent)

        self.data = DEFAULT_SETTINGS

        if os.path.isfile(CONFIG_FILE) and load_file:
            config_string = open(CONFIG_FILE, 'r').read()
            try:
                self.data = json.loads(config_string)
            except json.JSONDecodeError:
                log.error("Invalid configuration file. Falling back to defaults.")


    @pyqtSlot()
    def save(self):
        with open(CONFIG_FILE, 'w') as cfg:
            json.dump(self.data, cfg, indent=2)

    @pyqtSlot(bool)
    def set_minimize_on_record(self, value):
        self.data['minimize_on_record'] = value

    @pyqtSlot(float)
    def set_replay_delay(self, value):
        self.data['replay_delay'] = value

    @pyqtSlot(str)
    def append_to_history(self, value):
        self.data['history']['last_location'] = os.path.dirname(value)
        last_files = self.data['history']['last_files']

        if value in last_files:
            last_files.remove(value)

        last_files.insert(0, value)

        while len(last_files) > self.data['history']['max_size']:
            del last_files[-1]

    @pyqtSlot(str)
    def sel_last_dir(self, value):
        self.data['history']['last_location'] = value

SETTINGS = Settings()
CFG = SETTINGS.data



from datetime import datetime

from PyQt5.QtCore import Qt
from PyQt5.QtCore import QVariant
from PyQt5.QtWidgets import QStyledItemDelegate, QComboBox

from .properties import PropertiesTableModel
from .data.events import Event
from ..ui.modifier_picker import ModifierCheckboxes


class EventPropertiesModel(PropertiesTableModel):
    def __init__(self, event, parent=None):
        """
        :type event: EventList
        """
        super().__init__(event, parent)
        self.event_class_name = event.__class__.__name__

    def data(self, index, role=Qt.DisplayRole):
        """
        Returns the data for the specified cell for the specified role
        :param index: The table cell
        :type index: QModelIndex
        :param role: The Qt role of the returned value (only DisplayRole is implemented)
        """
        row = index.row()
        col = index.column()
        value = QVariant()
        if role in [Qt.DisplayRole, Qt.EditRole]:
            if row == 0:
                value = self.tr('type') if col == 0 else self.event_class_name
            else:
                props = self._dict.copy()
                if 'type' in props:
                    del props['type']

                if col == 0:
                    value = list(props.keys())[row-1]
                elif col == 1:
                    value = list(props.values())[row-1]
                    if isinstance(value, (list, tuple, set)):
                        if role != Qt.EditRole:
                            value = ','.join(value)
                    elif list(props.keys())[row-1] == 'time':
                        value = str(datetime.fromtimestamp(int(value)/1000))
        else:
            value = super().data(index, role)

        return value


class EventPropertiesItemDelegate(QStyledItemDelegate):
    """ Class handling the cell edition of the properties table """

    def createEditor(self, parent, option, index):
        """ Creating the widget used as cell editor """
        # ComboBox only in column 1
        if index.column() == 1 and index.row() == 0:
            # Create the combobox and populate it
            comboType = QComboBox(parent)
            comboType.addItems(list(map(lambda l: l.__name__, Event.get_subclasses())))
            # asserting self.parent is a Controller
            comboType.currentTextChanged.connect(self.parent().change_event_type)
            return comboType

        if index.column() == 1 and isinstance(index.data(Qt.EditRole), (list, tuple, set)):
            checkModifiers = ModifierCheckboxes(parent)
            checkModifiers.values_changed.connect(self.parent().set_shortcut_modifiers)
            return checkModifiers

        return super().createEditor(parent, option, index)

    def setEditorData(self, editor, index):
        """ Definfing the data according to the cell editor type """
        if isinstance(editor, QComboBox):
            # get the index of the text in the combobox that matches the current value of the itenm
            current_text = str(index.data(Qt.EditRole))
            cb_index = editor.findText(current_text)
            # if it is valid, adjust the combobox
            if cb_index >= 0:
                editor.blockSignals(True)
                editor.setCurrentIndex(cb_index)
                editor.blockSignals(False)
        elif isinstance(editor, ModifierCheckboxes):
            editor.blockSignals(True)
            editor.set_values(index.data(Qt.EditRole))
            editor.blockSignals(False)
        else:
            super().setEditorData(editor, index)

    def setModelData(self, editor, model, index):
        """ Sets the edited data to the model """
        if isinstance(editor, QComboBox):
            # save the current text of the combo box as the current value of the item
            model.setData(index, editor.currentText(), Qt.EditRole)
        elif isinstance(editor, ModifierCheckboxes):
            model.setData(index, editor.values(), Qt.EditRole)
        else:
            super().setModelData(editor, model, index)



"""
Module managing the validation screenshots.
"""
import logging as log
import time
import io
from tempfile import gettempdir

import pyautogui
import numpy
import PIL
from PIL.Image import Image
from PIL.ImageQt import ImageQt
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import QBuffer, QByteArray, qCompress, qUncompress
import cv2

def pixmap_to_base64(image):
    '''
    :type image: QPixmap
    :return: bytes
    '''
    if not image:
        log.error('Not an image')
        return None

    data = QByteArray()
    buf = QBuffer(data)
    image.save(buf, 'PNG')
    return qCompress(data).toBase64().data().decode('utf8')

def base64_to_pixmap(image):
    '''
    :type image: bytes
    :return: QPixmap
    '''
    img = QPixmap()
    if img.loadFromData(qUncompress(QByteArray.fromBase64(bytes(image, 'utf8'))), 'PNG'):
        return img

    log.error("Can't convert to image: %s", image)
    return None


class ImageConverter:
    """
    Class handling the different image formats used in the program:
      - QPixmap: used to display in the interface
      - numpy.ndarray (OpenCV): to compute differences between images
      - PIL.Image: format in which the screenshots are taken
    The constructor takes one of the three supported image formats,
    automatically detects which format is used, and converts it in the two
    others.
    """
    def __init__(self, image):
        self.pixmap = None
        self.opencv = None
        self.pilimg = None

        self.set_image(image)

    def set_image(self, image):
        if image is None:
            raise "Can't convert null image"

        if isinstance(image, Image):
            self.pilimg = image
            self._pil_to_opencv()
            self._pil_to_pixmap()
        elif isinstance(image, QPixmap):
            self.pixmap = image
            self._pixmap_to_pil()
            self._pil_to_opencv()
        elif isinstance(image, numpy.ndarray):
            self.opencv = image
            self._opencv_to_pil()
            self._pil_to_pixmap()
        else:
            raise "Unsupported image format: {}".format(type(image))

    def _pil_to_opencv(self):
        img = numpy.array(self.pilimg)
        self.opencv = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

    def _opencv_to_pil(self):
        img = cv2.cvtColor(self.opencv, cv2.COLOR_BGR2RGB)
        self.pilimg = PIL.Image.fromarray(img)

    def _pil_to_pixmap(self):
        self.pixmap = QPixmap.fromImage(ImageQt(self.pilimg).copy())

    def _pixmap_to_pil(self):
        buffer = QBuffer()
        buffer.open(QBuffer.ReadWrite)
        self.pixmap.save(buffer, "PNG")
        self.pilimg = PIL.Image.open(io.BytesIO(buffer.data()))


class Screenshot:
    """
    Data relative to screenshots and test case.
    Contains an "original" image to be used as reference and a "result" image
    acting as the final result of the test, the size and coords of the capture.
    """
    def __init__(self, region=()):
        # PNG as base64
        self.image = None
        self.result = None

        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0

        self.ocr_expected = ''
        self.ocr_image = ''
        self.ocr_result = ''
        self.ocr_characters = ''

        if region:
            self.x, self.y, self.w, self.h = region

    def region(self):
        """ Size and position of the screenshot """
        return (self.x, self.y, self.w, self.h)

    def do_capture(self, target='image'):
        """
        Performs screenshot at the location defined by region()
        :param target: 'image' or 'result' depending on the type of the image captured
        """
        path = '{0}/capture-{1}.png'.format(gettempdir(), round(time.time()*1000))
        pyautogui.screenshot(path, region=(self.x, self.y, self.w, self.h))
        log.info('Capture performed')

        encoded_image = pixmap_to_base64(QPixmap(path))
        if target == 'result':
            self.result = encoded_image
        else:
            self.image = encoded_image
        log.info('Capture converted')

    def as_image(self, target='image'):
        """
        :param target: 'image' or 'result' depending on the type of the image to convert
        :return: QPixmap of the snapshot
        """
        img = self.result if target == 'result' else self.image
        return base64_to_pixmap(img)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './snitch/ui/diff.ui'
#
# Created by: PyQt5 UI code generator 5.14.1
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Diff(object):
    def setupUi(self, Diff):
        Diff.setObjectName("Diff")
        Diff.resize(1094, 681)
        Diff.setAutoFillBackground(True)
        self.gridLayout = QtWidgets.QGridLayout(Diff)
        self.gridLayout.setObjectName("gridLayout")
        self.widget = QtWidgets.QWidget(Diff)
        self.widget.setObjectName("widget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.widget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(646, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.highlightButton = QtWidgets.QPushButton(self.widget)
        self.highlightButton.setCheckable(True)
        self.highlightButton.setObjectName("highlightButton")
        self.horizontalLayout.addWidget(self.highlightButton)
        self.pushButton = QtWidgets.QPushButton(self.widget)
        self.pushButton.setObjectName("pushButton")
        self.horizontalLayout.addWidget(self.pushButton)
        self.gridLayout.addWidget(self.widget, 0, 0, 1, 2)
        self.referenceGroupBox = QtWidgets.QGroupBox(Diff)
        self.referenceGroupBox.setObjectName("referenceGroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.referenceGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.referenceScrollArea = QtWidgets.QScrollArea(self.referenceGroupBox)
        self.referenceScrollArea.setWidgetResizable(True)
        self.referenceScrollArea.setObjectName("referenceScrollArea")
        self.referenceScrollAreaContents = QtWidgets.QWidget()
        self.referenceScrollAreaContents.setGeometry(QtCore.QRect(0, 0, 507, 466))
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        self.referenceScrollAreaContents.setPalette(palette)
        self.referenceScrollAreaContents.setObjectName("referenceScrollAreaContents")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.referenceScrollAreaContents)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.referenceImageLabel = QtWidgets.QLabel(self.referenceScrollAreaContents)
        self.referenceImageLabel.setText("")
        self.referenceImageLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.referenceImageLabel.setObjectName("referenceImageLabel")
        self.horizontalLayout_2.addWidget(self.referenceImageLabel)
        self.referenceScrollArea.setWidget(self.referenceScrollAreaContents)
        self.verticalLayout.addWidget(self.referenceScrollArea)
        self.referenceTextEdit = QtWidgets.QPlainTextEdit(self.referenceGroupBox)
        self.referenceTextEdit.setMaximumSize(QtCore.QSize(16777215, 120))
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.referenceTextEdit.setFont(font)
        self.referenceTextEdit.setUndoRedoEnabled(False)
        self.referenceTextEdit.setReadOnly(True)
        self.referenceTextEdit.setObjectName("referenceTextEdit")
        self.verticalLayout.addWidget(self.referenceTextEdit)
        self.gridLayout.addWidget(self.referenceGroupBox, 1, 0, 1, 1)
        self.groupBox = QtWidgets.QGroupBox(Diff)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.resultScrollArea = QtWidgets.QScrollArea(self.groupBox)
        self.resultScrollArea.setWidgetResizable(True)
        self.resultScrollArea.setObjectName("resultScrollArea")
        self.resultScrollAreaContents = QtWidgets.QWidget()
        self.resultScrollAreaContents.setGeometry(QtCore.QRect(0, 0, 507, 466))
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(51, 51, 51))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        self.resultScrollAreaContents.setPalette(palette)
        self.resultScrollAreaContents.setObjectName("resultScrollAreaContents")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.resultScrollAreaContents)
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_3.setSpacing(0)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.resultImageLabel = QtWidgets.QLabel(self.resultScrollAreaContents)
        self.resultImageLabel.setText("")
        self.resultImageLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.resultImageLabel.setObjectName("resultImageLabel")
        self.horizontalLayout_3.addWidget(self.resultImageLabel)
        self.resultScrollArea.setWidget(self.resultScrollAreaContents)
        self.verticalLayout_2.addWidget(self.resultScrollArea)
        self.resultTextEdit = QtWidgets.QPlainTextEdit(self.groupBox)
        self.resultTextEdit.setMaximumSize(QtCore.QSize(16777215, 120))
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.resultTextEdit.setFont(font)
        self.resultTextEdit.setUndoRedoEnabled(False)
        self.resultTextEdit.setReadOnly(True)
        self.resultTextEdit.setObjectName("resultTextEdit")
        self.verticalLayout_2.addWidget(self.resultTextEdit)
        self.gridLayout.addWidget(self.groupBox, 1, 1, 1, 1)

        self.retranslateUi(Diff)
        self.highlightButton.released.connect(Diff.swap_results)
        self.pushButton.released.connect(Diff.save_images)
        QtCore.QMetaObject.connectSlotsByName(Diff)

    def retranslateUi(self, Diff):
        _translate = QtCore.QCoreApplication.translate
        Diff.setWindowTitle(_translate("Diff", "Screenshot comparison"))
        self.highlightButton.setText(_translate("Diff", "Highlight differences"))
        self.pushButton.setText(_translate("Diff", "Export as image…"))
        self.referenceGroupBox.setTitle(_translate("Diff", "Reference"))
        self.groupBox.setTitle(_translate("Diff", "Test result"))
